"""Provides the code for saving and loading the history."""

##############################################################################
# Python imports.
from json import dumps, loads
from pathlib import Path

##############################################################################
# httpx imports.
from httpx import URL

##############################################################################
# Local imports.
from ..types import HikeHistory
from .locations import data_dir


##############################################################################
def history_file() -> Path:
    """Get the path for the history file.

    Returns:
        The path for the history file.
    """
    return data_dir() / "history.json"


##############################################################################
def save_history(history: HikeHistory) -> None:
    """Save the history to storage.

    Args:
        history: The history to save.
    """
    history_file().write_text(
        dumps(
            [
                ("url" if isinstance(entry, URL) else "path", str(entry))
                for entry in history
            ],
            indent=4,
        ),
        encoding="utf-8",
    )


##############################################################################
def load_history() -> HikeHistory:
    """Load the history from storage.

    Returns:
        The loaded history.
    """
    return HikeHistory(
        [
            (URL if entry_type == "url" else Path)(entry)
            for entry_type, entry in loads(history.read_text(encoding="utf-8"))
        ]
        if (history := history_file()).exists()
        else []
    )


### history.py ends here



"""Provides a simple fallback text editor for editing some Markdown."""

##############################################################################
# Python imports.
from pathlib import Path

##############################################################################
# Textual imports.
from textual import on, work
from textual.app import ComposeResult
from textual.reactive import var
from textual.widgets import Footer, Header, TextArea

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import ChangeTheme, Command, Help
from textual_enhanced.dialogs import Confirm
from textual_enhanced.screen import EnhancedScreen

##############################################################################
# Local imports.
from .commands import Close, EditorCommands, Save


##############################################################################
class Editor(EnhancedScreen[None]):
    """A fallback editor for editing some Markdown."""

    DEFAULT_CSS = """
    Editor {
        TextArea, TextArea:focus {
            border: none;
        }
    }
    """

    COMMAND_MESSAGES = (
        Help,
        Save,
        ChangeTheme,
        Close,
    )

    BINDINGS = Command.bindings(*COMMAND_MESSAGES)
    COMMANDS = {EditorCommands}

    HELP = """
    ## Markdown Editor

    This is a simple built-in editor for Markdown. If you wish to use your
    own editor set either the `$VISUAL` or `$EDITOR` environment variables
    to suit your taste.

    Commands available in the editor are:
    """

    _dirty: var[bool | None] = var(None)
    """Is the document dirty?"""

    def __init__(self, location: Path) -> None:
        """Initialise the screen.

        Args:
            location: The location to edit.
        """
        self._location = location
        """The location to edit."""
        super().__init__()
        self.title = str(location)

    def compose(self) -> ComposeResult:
        """Compose the screen."""
        yield Header()
        yield TextArea.code_editor(language="markdown")
        yield Footer()

    def on_mount(self) -> None:
        """Configure the screen when the DOM is mounted."""
        try:
            self.query_one(TextArea).text = self._location.read_text(encoding="utf-8")
        except IOError as error:
            self.notify(
                str(error),
                title=f"Error reading from {self._location}",
                severity="error",
                timeout=8,
            )

    def _watch__dirty(self) -> None:
        """Handle the dirty flag being modified."""
        self.sub_title = "Edited" if self._dirty else ""

    @on(TextArea.Changed)
    def _go_dirty(self) -> None:
        """Mark that the document is dirty."""
        # We'll get a Changed event on startup, so we start with None, then
        # go to False, then go to True otherwise.
        self._dirty = False if self._dirty is None else True

    @on(Save)
    def action_save_command(self) -> None:
        """Save the editor content."""
        try:
            self._location.write_text(self.query_one(TextArea).text, encoding="utf-8")
        except IOError as error:
            self.notify(
                str(error),
                title=f"Error writing to {self._location}",
                severity="error",
                timeout=8,
            )
            return
        self._dirty = False

    @on(Close)
    @work
    async def action_close_command(self) -> None:
        """Close the editor."""
        if self._dirty:
            if not await self.app.push_screen_wait(
                Confirm(
                    "Unsaved changes",
                    "You have unsaved changed in your document. Are you sure you want to quit?",
                )
            ):
                return
        self.dismiss()


### screen.py ends here



"""The main screen for the application."""

##############################################################################
# Python imports.
from argparse import Namespace
from functools import partial

##############################################################################
# Pyperclip imports.
from pyperclip import PyperclipException
from pyperclip import copy as copy_to_clipboard

##############################################################################
# Textual imports.
from textual import on, work
from textual.app import ComposeResult
from textual.containers import Horizontal, VerticalGroup
from textual.reactive import var
from textual.widgets import Footer, Header, Markdown

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import ChangeTheme, Command, Help, Quit
from textual_enhanced.dialogs import ModalInput
from textual_enhanced.screen import EnhancedScreen
from textual_enhanced.tools import add_key

##############################################################################
# Textual fspicker imports.
from textual_fspicker import FileOpen, FileSave, Filters

##############################################################################
# Local imports.
from .. import __version__
from ..commands import (
    Backward,
    BookmarkLocation,
    ChangeCommandLineLocation,
    ChangeNavigationSide,
    CopyLocationToClipboard,
    CopyMarkdownToClipboard,
    Edit,
    Forward,
    JumpToBookmarks,
    JumpToCommandLine,
    JumpToDocument,
    JumpToHistory,
    JumpToLocalBrowser,
    JumpToTableOfContents,
    Reload,
    SaveCopy,
    SearchBookmarks,
    ToggleNavigation,
)
from ..data import (
    load_bookmarks,
    load_command_history,
    load_configuration,
    load_history,
    maybe_markdown,
    save_bookmarks,
    save_command_history,
    save_history,
    update_configuration,
)
from ..messages import (
    ClearHistory,
    CopyToClipboard,
    HandleInput,
    OpenFrom,
    OpenFromForge,
    OpenFromHistory,
    OpenLocation,
    RemoveHistoryEntry,
    SetLocalViewRoot,
)
from ..providers import BookmarkCommands, MainCommands
from ..support import view_in_browser
from ..widgets import CommandLine, Navigation, Viewer


##############################################################################
class Main(EnhancedScreen[None]):
    """The main screen for the application."""

    TITLE = f"Hike v{__version__}"

    DEFAULT_CSS = """
    Main {
        #workspace {
            hatch: right $surface;
            .panel {
                border-left: solid $panel;
                &:focus, &:focus-within {
                    border-left: solid $border;
                }
            }
        }

        .panel {
            background: $surface;
            &:focus-within {
                background: $panel 80%;
            }
            * {
                scrollbar-background: $surface;
                scrollbar-background-hover: $surface;
                scrollbar-background-active: $surface;
            }
            &:focus-within * {
                scrollbar-background: $panel;
                scrollbar-background-hover: $panel;
                scrollbar-background-active: $panel;
            }
        }

        Navigation {
            display: none;
        }
        &.navigation Navigation {
            display: block;
        }
    }
    """

    HELP = """
    ## Main application keys and commands

    The following key bindings and commands are available:
    """

    COMMAND_MESSAGES = (
        # Keep these together as they're bound to function keys and destined
        # for the footer.
        Help,
        ToggleNavigation,
        Edit,
        ChangeTheme,
        Quit,
        # Everything else.
        Backward,
        BookmarkLocation,
        ChangeCommandLineLocation,
        ChangeNavigationSide,
        CopyLocationToClipboard,
        CopyMarkdownToClipboard,
        Forward,
        JumpToBookmarks,
        JumpToCommandLine,
        JumpToDocument,
        JumpToHistory,
        JumpToLocalBrowser,
        JumpToTableOfContents,
        Reload,
        SaveCopy,
        SearchBookmarks,
    )

    BINDINGS = Command.bindings(*COMMAND_MESSAGES)
    COMMANDS = {MainCommands}

    AUTO_FOCUS = "CommandLine Input"

    navigation_visible: var[bool] = var(True)
    """Set if the navigation panel is visible or not."""

    def __init__(self, arguments: Namespace) -> None:
        """Initialise the main screen.

        Args:
            arguments: The arguments passed to the application on the command line.
        """
        self._arguments = arguments
        """The arguments passed on the command line."""
        super().__init__()

    def compose(self) -> ComposeResult:
        """Compose the content of the screen."""
        yield Header()
        with VerticalGroup():
            with Horizontal(id="workspace"):
                yield Navigation(classes="panel")
                yield Viewer(classes="panel")
            yield CommandLine(classes="panel")
        yield Footer()

    def on_mount(self) -> None:
        """Configure the screen once the DOM is mounted."""
        config = load_configuration()
        self.navigation_visible = config.navigation_visible
        self.query_one(Navigation).dock_right = config.navigation_on_right
        self.query_one(Navigation).bookmarks = (bookmarks := load_bookmarks())
        BookmarkCommands.bookmarks = bookmarks
        self.query_one(Viewer).history = load_history()
        self.query_one(CommandLine).history = load_command_history()
        self.query_one(CommandLine).dock_top = config.command_line_on_top
        if self._arguments.command:
            self.post_message(HandleInput(" ".join(self._arguments.command)))

    def _watch_navigation_visible(self) -> None:
        """React to the navigation visible property being set."""
        self.set_class(self.navigation_visible, "navigation")
        with update_configuration() as config:
            config.navigation_visible = self.navigation_visible

    @on(HandleInput)
    def _handle_input(self, message: HandleInput) -> None:
        """Handle input as if the user had typed it in.

        Args:
            message: The message requesting the input be handled.
        """
        self.query_one(CommandLine).handle_input(message.user_input)

    @on(CopyToClipboard)
    def _copy_text_to_clipbaord(self, message: CopyToClipboard) -> None:
        """Copy some text into the clipboard.

        Args:
            message: The message requesting the text be copied.
        """
        # First off, use Textual's own copy to clipboard facility. Generally
        # this will work in most terminals, and if it does it'll likely work
        # best, getting the text through remote connections to the user's
        # own environment.
        self.app.copy_to_clipboard(message.text)
        # However, as a backup, use pyerclip too. If the above did fail due
        # to the terminal not supporting the operation, this might.
        try:
            copy_to_clipboard(message.text)
        except PyperclipException:
            pass
        # Give the user some feedback.
        self.notify("Copied")

    @on(OpenLocation)
    def _open_markdown(self, message: OpenLocation) -> None:
        """Open a file for viewing.

        Args:
            message: The message requesting the file be opened.
        """
        if maybe_markdown(message.to_open):
            self.query_one(Viewer).location = message.to_open
        else:
            view_in_browser(message.to_open)

    @on(OpenFrom)
    @work
    async def _browse_for_file(self, message: OpenFrom) -> None:
        """Browse for a markdown file with a file open dialog.

        Args:
            message: The message requesting the operation.
        """
        if chosen := await self.app.push_screen_wait(
            FileOpen(
                message.location,
                filters=Filters(
                    (
                        "Markdown",
                        maybe_markdown,
                    ),
                    ("All files", lambda _: True),
                ),
                cancel_button=partial(add_key, key="Esc", context=self),
            )
        ):
            self.post_message(OpenLocation(chosen))

    @on(OpenFromHistory)
    def _open_from_history(self, message: OpenFromHistory) -> None:
        """Open a location from the history.

        Args:
            message: The message requesting the history open.
        """
        self.query_one(Viewer).goto(message.location)

    @on(OpenFromForge)
    @work
    async def _open_from_forge(self, message: OpenFromForge) -> None:
        """Open a file from a git forge.

        Args:
            message: The message requesting the operation.
        """
        if (url := await message.url()) is None:
            self.notify(
                "The file you were after could not be located.\n\n"
                "Check the spelling of the owner, repository and file; "
                "also consider what branch it might be on.",
                title=f"No suitable {message.forge} target found",
                severity="error",
                timeout=8,
            )
        else:
            self.post_message(OpenLocation(url))

    @on(RemoveHistoryEntry)
    def _remove_location_from_history(self, message: RemoveHistoryEntry) -> None:
        """Remove a specific location from history.

        Args:
            message: The message requesting the location be removed.
        """
        self.query_one(Viewer).remove_from_history(message.location)

    @on(ClearHistory)
    def _clear_down_history(self) -> None:
        """Clear all items from history."""
        self.query_one(Viewer).clear_history()

    @on(SetLocalViewRoot)
    def _set_local_root(self, message: SetLocalViewRoot) -> None:
        """Change the root directory of the local file browser.

        Args:
            message: The message requesting the root be changed.
        """
        self.query_one(Navigation).set_local_view_root(message.root)

    @on(Markdown.TableOfContentsUpdated)
    def _update_navigation_contents(
        self, message: Markdown.TableOfContentsUpdated
    ) -> None:
        """Handle the table of contents being updated.

        Args:
            message: The message broadcasting that the ToC is updated.
        """
        self.query_one(Navigation).table_of_contents = message.table_of_contents

    @on(Markdown.TableOfContentsSelected)
    def _jump_to_content(self, message: Markdown.TableOfContentsSelected) -> None:
        """Jump to a specific location in the current document.

        Args:
            message: The message request the jump.
        """
        self.query_one(Viewer).jump_to_content(message.block_id)

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        """Check if an action is possible to perform right now.

        Args:
            action: The action to perform.
            parameters: The parameters of the action.

        Returns:
            `True` if it can perform, `False` or `None` if not.
        """
        if not self.is_mounted:
            # Surprisingly it seems that Textual's "dynamic bindings" can
            # cause this method to be called before the DOM is up and
            # running. This breaks the rule of least astonishment, I'd say,
            # but okay let's be defensive... (when I can come up with a nice
            # little MRE I'll report it).
            return True
        if action == Forward.action_name():
            return self.query_one(Viewer).history.can_go_forward or None
        if action == Backward.action_name():
            return self.query_one(Viewer).history.can_go_backward or None
        if action == Edit.action_name():
            return self.query_one(Viewer).is_editable or None
        if action in (
            command.action_name()
            for command in (
                BookmarkLocation,
                CopyLocationToClipboard,
                CopyMarkdownToClipboard,
                Reload,
                SaveCopy,
            )
        ):
            return self.query_one(Viewer).location is not None
        return True

    @on(Reload)
    def action_reload_command(self) -> None:
        """Reload the current document."""
        self.query_one(Viewer).reload()

    @on(SearchBookmarks)
    def action_search_bookmarks_command(self) -> None:
        """Search the bookmarks in the command palette."""
        self.show_palette(BookmarkCommands)

    @on(ToggleNavigation)
    def action_toggle_navigation_command(self) -> None:
        """Toggle the display of the navigation panel."""
        self.navigation_visible = not self.navigation_visible

    @on(ChangeNavigationSide)
    def action_change_navigation_side_command(self) -> None:
        """Change the side that the navigation panel lives on."""
        navigation = self.query_one(Navigation)
        navigation.dock_right = not navigation.dock_right
        with update_configuration() as config:
            config.navigation_on_right = navigation.dock_right

    @on(ChangeCommandLineLocation)
    def action_change_command_line_location_command(self) -> None:
        """Change the location of the command line."""
        command_line = self.query_one(CommandLine)
        command_line.dock_top = not command_line.dock_top
        with update_configuration() as config:
            config.command_line_on_top = command_line.dock_top

    @on(JumpToCommandLine)
    def action_jump_to_command_line_command(self) -> None:
        """Jump to the command line."""
        if self.AUTO_FOCUS:
            self.query_one(self.AUTO_FOCUS).focus()

    @on(JumpToDocument)
    def action_jump_to_document_command(self) -> None:
        """Jump to the document."""
        self.query_one(Viewer).focus()

    @on(Backward)
    def action_backward_command(self) -> None:
        """Move backward through history."""
        self.query_one(Viewer).backward()

    @on(Forward)
    def action_forward_command(self) -> None:
        """Move forward through history."""
        self.query_one(Viewer).forward()

    @on(BookmarkLocation)
    @work
    async def action_bookmark_location_command(self) -> None:
        """Add the current location to the bookmarks."""
        if (location := self.query_one(Viewer).location) is None:
            self.notify(
                "Please visit a file or URL to bookmark it.", severity="warning"
            )
            return
        if location in self.query_one(Navigation).bookmarks:
            self.notify(
                "This location is already in your bookmarks", severity="warning"
            )
            return
        if title := await self.app.push_screen_wait(
            ModalInput("Enter a title for the bookmark")
        ):
            self.query_one(Navigation).add_bookmark(title, location)
            self.notify("Bookmark added")

    def _with_navigation_visible(self) -> Navigation:
        """Ensure navigation is visible.

        Returns:
            The navigation widget.
        """
        self.navigation_visible = True
        return self.query_one(Navigation)

    @on(JumpToTableOfContents)
    def action_jump_to_table_of_contents_command(self) -> None:
        """Jump to the table of contents."""
        self._with_navigation_visible().jump_to_content()

    @on(JumpToLocalBrowser)
    def action_jump_to_local_browser_command(self) -> None:
        """Jump to the local browser."""
        self._with_navigation_visible().jump_to_local()

    @on(JumpToBookmarks)
    def action_jump_to_bookmarks_command(self) -> None:
        """Jump to the bookmarks."""
        self._with_navigation_visible().jump_to_bookmarks()

    @on(JumpToHistory)
    def action_jump_to_history_command(self) -> None:
        """Jump to the history."""
        self._with_navigation_visible().jump_to_history()

    @on(CopyLocationToClipboard)
    def action_copy_location_to_clipboard_command(self) -> None:
        """Copy the current location to the clipboard."""
        self.post_message(CopyToClipboard(str(self.query_one(Viewer).location)))

    @on(CopyMarkdownToClipboard)
    def action_copy_markdown_to_clipboard_command(self) -> None:
        """Copy the current markdown to the clipboard."""
        self.post_message(CopyToClipboard(self.query_one(Viewer).source))

    @on(Edit)
    def action_edit_command(self) -> None:
        """Edit the current markdown document, if possible."""
        self.query_one(Viewer).edit()

    @on(SaveCopy)
    @work
    async def action_save_copy_command(self) -> None:
        """Save a copy of the current document to a new file."""
        if (suggested := self.query_one(Viewer).filename) is None:
            return
        if save_to := await self.app.push_screen_wait(
            FileSave(
                default_file=suggested,
                cancel_button=partial(add_key, key="Esc", context=self),
            )
        ):
            try:
                save_to.write_text(self.query_one(Viewer).source, encoding="utf-8")
            except IOError as error:
                self.notify(str(error), title="Save Error", severity="error", timeout=8)
                return
            self.notify(f"Saved {save_to}")
            self.query_one(Navigation).refresh_local_view()

    @on(Viewer.HistoryUpdated)
    def _update_history(self, message: Viewer.HistoryUpdated) -> None:
        """Update the view of history when it changes.

        Args:
            message: The message to say that history changed.
        """
        self.query_one(Navigation).update_history(message.viewer.history)
        save_history(message.viewer.history)

    @on(Viewer.HistoryVisit)
    def _move_history(self, message: Viewer.HistoryVisit) -> None:
        """React to a new location in history being visited."""
        if (location := message.viewer.history.current_location) is not None:
            self.query_one(Navigation).highlight_history(location)

    @on(Navigation.BookmarksUpdated)
    def _update_bookmarks(self, message: Navigation.BookmarksUpdated) -> None:
        """Handle the bookmarks being updated."""
        save_bookmarks(message.navigation.bookmarks)
        BookmarkCommands.bookmarks = message.navigation.bookmarks

    @on(CommandLine.HistoryUpdated)
    def _save_command_line_history(self, message: CommandLine.HistoryUpdated) -> None:
        """Save the command line history to storage when it changes.

        Args:
            message: The message requesting the same.
        """
        save_command_history(message.command_line.history)


### main.py ends here



"""Provides a command for browsing Obsidian vaults."""

##############################################################################
# Python imports.
from pathlib import Path

##############################################################################
# Textual imports.
from textual.widget import Widget

##############################################################################
# Local imports.
from ...data import load_configuration
from ...messages import SetLocalViewRoot
from .base_command import InputCommand


##############################################################################
class ObsidianCommand(InputCommand):
    """Change the root directory to your Obsidian vaults"""

    COMMAND = "`obsidian`"
    ALIASES = "`obs`"

    @classmethod
    def handle(cls, text: str, for_widget: Widget) -> bool:
        """Handle the command.

        Args:
            text: The text of the command.
            for_widget: The widget to handle the command for.

        Returns:
            `True` if the command was handled; `False` if not.
        """
        if cls.is_command(text):
            if vaults := Path(load_configuration().obsidian_vaults).expanduser():
                if vaults.exists() and vaults.is_dir():
                    for_widget.post_message(SetLocalViewRoot(vaults.resolve()))
                    return True
        return False


### obsidian.py ends here



"""The main application class."""

##############################################################################
# Python imports.
from argparse import Namespace

##############################################################################
# Textual imports.
from textual.app import InvalidThemeError

##############################################################################
# Textual enhanced imports.
from textual_enhanced.app import EnhancedApp

##############################################################################
# Local imports.
from . import __version__
from .data import (
    load_configuration,
    update_configuration,
)
from .screens import Main


##############################################################################
class Hike(EnhancedApp[None]):
    """The main application class."""

    HELP_TITLE = f"Hike {__version__}"
    HELP_ABOUT = """
    `Hike` is a terminal-based Markdown viewer; it was created
    by and is maintained by [Dave Pearson](https://www.davep.org/); it is
    Free Software and can be [found on
    GitHub](https://github.com/davep/hike).
    """
    HELP_LICENSE = """
    Hike - A Markdown viewer for the terminal.  \n    Copyright (C) 2025 Dave Pearson

    This program is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program. If not, see <https://www.gnu.org/licenses/>.
    """

    COMMANDS = set()

    def __init__(self, arguments: Namespace) -> None:
        """Initialise the application.

        Args:
            The command line arguments passed to the application.
        """
        self._arguments = arguments
        """The command line arguments passed to the application."""
        super().__init__()
        configuration = load_configuration()
        if configuration.theme is not None:
            try:
                self.theme = configuration.theme
            except InvalidThemeError:
                pass
        self.update_keymap(configuration.bindings)

    def watch_theme(self) -> None:
        """Save the application's theme when it's changed."""
        with update_configuration() as config:
            config.theme = self.theme

    def get_default_screen(self) -> Main:
        """Get the default screen for the application.

        Returns:
            The main screen.
        """
        return Main(self._arguments)


### hike.py ends here



"""Provides the navigation panel widget."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from dataclasses import dataclass
from pathlib import Path

##############################################################################
# Textual imports.
from textual import on
from textual.app import ComposeResult
from textual.containers import Vertical
from textual.message import Message
from textual.reactive import var
from textual.widgets import Markdown, TabbedContent, TabPane, Tabs, Tree
from textual.widgets.markdown import MarkdownTableOfContents, TableOfContentsType

##############################################################################
# Textual enhanced imports.
from textual_enhanced.binding import HelpfulBinding

##############################################################################
# Local imports.
from ...commands import JumpToCommandLine
from ...data import Bookmark, Bookmarks, load_configuration
from ...types import HikeHistory, HikeLocation
from .bookmarks_view import BookmarksView
from .history_view import HistoryView
from .local_view import LocalView


##############################################################################
class Navigation(Vertical):
    """The navigation panel."""

    DEFAULT_CSS = """
    Navigation {
        width: 27%;
        min-width: 38;
        dock: left;
        background: transparent;

        &.--dock-right {
            dock: right;
        }

        #tabs-list {
            background: $panel;
        }

        /* https://github.com/Textualize/textual/issues/5488 */
        MarkdownTableOfContents, &:focus-within MarkdownTableOfContents {
            background: transparent;
            width: 1fr;
            Tree {
                background: transparent;
            }
        }

        /* https://github.com/Textualize/textual/issues/5488 */
        HistoryView, &:focus-within HistoryView,
        LocalView, &:focus-within LocalView,
        BookmarksView, &:focus-within BookmarksView {
            background: transparent;
        }
    }
    """

    BINDINGS = [
        ("escape", "return_to_tabs_or_bounce_out"),
        ("down", "move_into_panel"),
        HelpfulBinding(
            "ctrl+left",
            "switch('previous_tab')",
            tooltip="Move to the previous navigation tab",
        ),
        HelpfulBinding(
            "ctrl+right",
            "switch('next_tab')",
            tooltip="Move to the next navigation tab",
        ),
    ]

    HELP = """
    ## The Navigation Panel

    Here you'll find panels for the current document's table of contents; a
    local file browser (use the `chdir` command in the command line at the
    bottom of the screen to change the root directory); a simple bookmark
    manager and your browsing history.

    ### Useful keys
    """

    dock_right: var[bool] = var(False)
    """Should the navigation dock to the right?"""

    table_of_contents: var[TableOfContentsType | None] = var(None)
    """The currently-displayed table of contents."""

    bookmarks: var[Bookmarks] = var(Bookmarks)
    """The bookmarks."""

    def action_return_to_tabs_or_bounce_out(self) -> None:
        """Return focus to the tabs, or to the input."""
        if self.screen.focused == (tabs := self.query_one(Tabs)):
            self.post_message(JumpToCommandLine())
        else:
            tabs.focus()

    def action_move_into_panel(self) -> None:
        """Drop focus down into a panel."""
        if (active := self.query_one(TabbedContent).active_pane) is not None:
            for widget in active.query("*"):
                if widget.can_focus:
                    widget.focus()
                    return

    async def action_switch(self, switcher: str) -> None:
        await self.query_one(Tabs).run_action(switcher)
        self.call_after_refresh(self.run_action, "move_into_panel")

    def _watch_dock_right(self) -> None:
        """React to the dock toggle being changed."""
        self.set_class(self.dock_right, "--dock-right")

    def _maybe_enable_tab(self, tab: str, data: object) -> bool:
        """Enable/disable a tab based on there being data.

        Args:
            tab: The name of the tab.
            data: The data to test.

        Returns:
            `True` if the tab was enabled, `False` if disabled.
        """
        tabs = self.query_one(TabbedContent)
        if data:
            tabs.enable_tab(tab)
            return True
        tabs.disable_tab(tab)
        if tabs.active == tab:
            tabs.active = "local"
        return False

    def _watch_table_of_contents(self) -> None:
        """React to the table of content being updated."""
        self.query_one(
            MarkdownTableOfContents
        ).table_of_contents = self.table_of_contents
        if self._maybe_enable_tab("content", self.table_of_contents):
            self.query_one("MarkdownTableOfContents Tree", Tree).cursor_line = 0

    def _watch_bookmarks(self) -> None:
        """React to the bookmarks being changed."""
        self.query_one(BookmarksView).update(self.bookmarks)
        self._maybe_enable_tab("bookmarks", self.bookmarks)

    def compose(self) -> ComposeResult:
        """Compose the content of the widget."""
        with TabbedContent():
            with TabPane("Content", id="content"):
                yield MarkdownTableOfContents(Markdown())
            with TabPane("Local", id="local"):
                yield LocalView(
                    Path(load_configuration().local_start_location).expanduser()
                )
            with TabPane("Bookmarks", id="bookmarks"):
                yield BookmarksView()
            with TabPane("History", id="history"):
                yield HistoryView()

    def update_history(self, history: HikeHistory) -> None:
        """Update the history display.

        Args:
            history: The history to display.
        """
        self.query_one(HistoryView).update(history)
        self._maybe_enable_tab("history", history)

    def highlight_history(self, history: int) -> None:
        """Highlight a specific entry in history.

        Args:
            The ID of the item of history to highlight.
        """
        self.query_one(HistoryView).highlight_location(history)

    def set_local_view_root(self, root: Path) -> None:
        """Set the root directory for the local file browser.

        Args:
            root: The new root directory.
        """
        self.query_one(LocalView).path = root

    def refresh_local_view(self) -> None:
        """Refresh the local view."""
        self.query_one(LocalView).reload()

    @dataclass
    class BookmarksUpdated(Message):
        """Message sent when the bookmarks are updated."""

        navigation: Navigation
        """The navigation widget sending the message."""

    def add_bookmark(self, title: str, location: HikeLocation) -> None:
        """Add a bookmark.

        Args:
            title: The title of the bookmark.
            location: The location to bookmark.
        """
        self.bookmarks = [Bookmark(title, location), *self.bookmarks]
        self.post_message(self.BookmarksUpdated(self))

    @on(BookmarksView.Renamed)
    def _bookmark_renamed(self, message: BookmarksView.Renamed) -> None:
        """Handle a bookmark being renamed.

        Args:
            message: The message requesting the deletion.
        """
        try:
            bookmark = self.bookmarks.index(message.renamed_from)
        except ValueError:
            self.notify(
                "Could not find the bookmark to modify it",
                title="Bookmark error",
                severity="error",
            )
            return
        (new_bookmarks := self.bookmarks.copy())[bookmark] = message.renamed_to
        self.bookmarks = new_bookmarks
        self.post_message(self.BookmarksUpdated(self))

    @on(BookmarksView.Deleted)
    def _bookmark_deleted(self, message: BookmarksView.Deleted) -> None:
        """handle a bookmark being deleted.

        Args:
            message: The message requesting the deletion.
        """
        try:
            bookmark = self.bookmarks.index(message.deleted)
        except ValueError:
            self.notify(
                "Could not find the bookmark to delete it",
                title="Bookmark error",
                severity="error",
            )
            return
        del (new_bookmarks := self.bookmarks.copy())[bookmark]
        self.bookmarks = new_bookmarks
        self.post_message(self.BookmarksUpdated(self))

    def _activate(self, panel: str) -> None:
        self.query_one(TabbedContent).active = panel
        self.call_next(self.run_action, "move_into_panel")

    def jump_to_content(self) -> None:
        """Jump into the content panel, if possible."""
        if self.table_of_contents:
            self._activate("content")

    def jump_to_local(self) -> None:
        """Jump into the local browser panel, if possible."""
        self._activate("local")

    def jump_to_bookmarks(self) -> None:
        """Jump into the bookmarks panel, if possible."""
        if self.bookmarks:
            self._activate("bookmarks")

    def jump_to_history(self) -> None:
        """Jump into the history panel, if possible."""
        if self.query_one(HistoryView).option_count:
            self._activate("history")


### navigation.py ends here



"""Provides the command for opening a file."""

##############################################################################
# Python imports.
from pathlib import Path

##############################################################################
# Textual imports.
from textual.widget import Widget

##############################################################################
# Local imports.
from ...messages import OpenLocation
from .base_command import InputCommand


##############################################################################
class OpenFileCommand(InputCommand):
    """View the file at `<file>`"""

    COMMAND = "`<file>`"

    @classmethod
    def handle(cls, text: str, for_widget: Widget) -> bool:
        """Handle the command.

        Args:
            text: The text of the command.
            for_widget: The widget to handle the command for.

        Returns:
            `True` if the command was handled; `False` if not.
        """
        if (path := Path(text).expanduser()).is_file():
            for_widget.post_message(OpenLocation(path.resolve()))
            return True
        return False


### open_directory.py ends here



"""Hike -- A Markdown viewer for the terminal."""

##############################################################################
# Python imports.
from importlib.metadata import version
from typing import Final

##############################################################################
# Main app information.
__author__ = "Dave Pearson"
__copyright__ = "Copyright 2025, Dave Pearson"
__credits__ = ["Dave Pearson"]
__maintainer__ = "Dave Pearson"
__email__ = "davep@davep.org"
__version__ = version("hike")
__licence__ = "GPLv3+"

##############################################################################
USER_AGENT: Final[str] = f"Hike v{__version__} (https://github.com/davep/hike)"
"""The user agent string for the viewer."""

### __init__.py ends here



"""The main entry point for the application."""

##############################################################################
# Python imports.
from argparse import ArgumentParser, Namespace
from inspect import cleandoc
from operator import attrgetter

##############################################################################
# Local imports.
from . import __doc__, __version__
from .hike import Hike


##############################################################################
def get_args() -> Namespace:
    """Get the command line arguments.

    Returns:
        The arguments.
    """

    # Build the parser.
    parser = ArgumentParser(
        prog="hike",
        description=__doc__,
        epilog=f"v{__version__}",
    )

    # Add --version
    parser.add_argument(
        "-v",
        "--version",
        help="Show version information",
        action="version",
        version=f"%(prog)s v{__version__}",
    )

    # Add --license
    parser.add_argument(
        "--license",
        "--licence",
        help="Show license information",
        action="store_true",
    )

    # Add --bindings
    parser.add_argument(
        "-b",
        "--bindings",
        help="List commands that can have their bindings changed",
        action="store_true",
    )

    # The remainder is going to be the initial command.
    parser.add_argument(
        "command",
        help="The initial command; can be any valid input to Hike's command line.",
        nargs="*",
    )

    # Finally, parse the command line.
    return parser.parse_args()


##############################################################################
def show_bindable_commands() -> None:
    """Show the commands that can have bindings applied."""
    from rich.console import Console
    from rich.markup import escape

    from .screens import Main

    console = Console(highlight=False)
    for command in sorted(Main.COMMAND_MESSAGES, key=attrgetter("__name__")):
        if command().has_binding:
            console.print(
                f"[bold]{escape(command.__name__)}[/] [dim italic]- {escape(command.tooltip())}[/]"
            )
            console.print(
                f"    [dim italic]Default: {escape(command.binding().key)}[/]"
            )


##############################################################################
def main() -> None:
    """The main entry point."""
    args = get_args()
    if args.license:
        print(cleandoc(Hike.HELP_LICENSE))
    elif args.bindings:
        show_bindable_commands()
    else:
        Hike(args).run()


##############################################################################
if __name__ == "__main__":
    main()


### __main__.py ends here



"""Provides the main application commands for the command palette."""

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import (
    ChangeTheme,
    CommandHits,
    CommandsProvider,
    Help,
    Quit,
)

##############################################################################
# Local imports.
from ..commands import (
    Backward,
    BookmarkLocation,
    ChangeCommandLineLocation,
    ChangeNavigationSide,
    CopyLocationToClipboard,
    CopyMarkdownToClipboard,
    Edit,
    Forward,
    JumpToBookmarks,
    JumpToCommandLine,
    JumpToDocument,
    JumpToHistory,
    JumpToLocalBrowser,
    JumpToTableOfContents,
    Reload,
    SaveCopy,
    SearchBookmarks,
    ToggleNavigation,
)


##############################################################################
class MainCommands(CommandsProvider):
    """Provides some top-level commands for the application."""

    def commands(self) -> CommandHits:
        """Provide the main application commands for the command palette.

        Yields:
            The commands for the command palette.
        """
        yield from self.maybe(Backward)
        yield from self.maybe(BookmarkLocation)
        yield ChangeCommandLineLocation()
        yield ChangeNavigationSide()
        yield ChangeTheme()
        yield from self.maybe(CopyLocationToClipboard)
        yield from self.maybe(CopyMarkdownToClipboard)
        yield from self.maybe(Edit)
        yield from self.maybe(Forward)
        yield Help()
        yield JumpToBookmarks()
        yield JumpToCommandLine()
        yield JumpToDocument()
        yield JumpToHistory()
        yield JumpToLocalBrowser()
        yield JumpToTableOfContents()
        yield Quit()
        yield from self.maybe(Reload)
        yield from self.maybe(SaveCopy)
        yield SearchBookmarks()
        yield ToggleNavigation()


### main.py ends here



"""Provides the command for opening a file from a forge."""

##############################################################################
# Python imports.
from re import Pattern, compile
from typing import Final

##############################################################################
# Textual imports.
from textual.widget import Widget

##############################################################################
# Local imports.
from ...data import load_configuration
from ...messages import OpenFromForge
from .base_command import InputCommand


##############################################################################
class OpenFromForgeCommand(InputCommand):
    """Base class for commands that open a file from a forge."""

    ARGUMENTS = "`<remote-file>`¹"

    WITHOUT_BRANCH: Final[Pattern[str]] = compile(
        r"^(?P<owner>[^/ ]+)[/ ](?P<repo>[^ :]+)(?: +(?P<file>[^ ]+))?$"
    )
    """Regular expression for finding forge details without a branch."""

    WITH_BRANCH: Final[Pattern[str]] = compile(
        r"^(?P<owner>[^/ ]+)[/ ](?P<repo>[^ :]+):(?P<branch>[^ ]+)(?: +(?P<file>[^ ]+))?$"
    )
    """Regular expression for finding forge details with a branch."""

    FORGE = ""
    """The name of the forge."""

    URL_FORMAT = ""
    """The format of the raw URL for the forge."""

    HELP = f"""
    | Format | Effect |
    | -- | -- |
    | `<owner>/<repo>` | Open `README.md` from a repository |
    | `<owner> <repo>` | Open `README.md` from a repository |
    | `<owner>/<repo> <file>` | Open a specific file from a repository |
    | `<owner> <repo> <file>` | Open a specific file from a repository |
    | `<owner>/<repo>:<branch>` | Open `README.md` from a specific branch of a repository |
    | `<owner> <repo>:<branch>` | Open `README.md` from a specific branch of a repository |
    | `<owner>/<repo>:<branch> <file>` | Open a specific file from a specific branch of a repository |
    | `<owner> <repo>:<branch> <file>` | Open a specific file from a specific branch of a repository |

    If `<branch>` is omitted the requested file is looked in the following branches:
    {", ".join(f"`{branch}`" for branch in load_configuration().main_branches)}.
    """

    @classmethod
    def maybe_request(cls, arguments: str, for_widget: Widget) -> bool:
        """Maybe request a file be opened from the given forge.

        Args:
            arguments: The arguments to parse.
            for_widget: The widget to send the request to.

        Returns:
            `True` if the arguments could be parsed, `False` if not.
        """
        if details := cls.WITHOUT_BRANCH.match(arguments):
            for_widget.post_message(
                OpenFromForge(
                    cls.FORGE,
                    cls.URL_FORMAT,
                    details["owner"],
                    details["repo"],
                    filename=details["file"],
                )
            )
            return True
        if details := cls.WITH_BRANCH.match(arguments):
            for_widget.post_message(
                OpenFromForge(
                    cls.FORGE,
                    cls.URL_FORMAT,
                    details["owner"],
                    details["repo"],
                    details["branch"],
                    details["file"],
                )
            )
            return True
        return False

    @classmethod
    def handle(cls, text: str, for_widget: Widget) -> bool:
        """Handle the forge command.

        Args:
            text: The text of the command.
            for_widget: The widget to handle the command for.

        Returns:
            `True` if the command was handled; `False` if not.
        """
        command, arguments = cls.split_command(text)
        return cls.is_command(command) and cls.maybe_request(arguments, for_widget)


##############################################################################
class OpenFromBitbucket(OpenFromForgeCommand):
    """Open a file from Bitbucket"""

    COMMAND = "`bitbucket`"
    ALIASES = "`bb`"
    FORGE = "Bitbucket"
    URL_FORMAT = "https://bitbucket.org/{owner}/{repository}/raw/{branch}/{file}"


##############################################################################
class OpenFromCodeberg(OpenFromForgeCommand):
    """Open a file from Codeberg"""

    COMMAND = "`codeberg`"
    ALIASES = "`cb`"
    FORGE = "Codeberg"
    URL_FORMAT = "https://codeberg.org/{owner}/{repository}/raw//branch/{branch}/{file}"


##############################################################################
class OpenFromGitHub(OpenFromForgeCommand):
    """Open a file from GitHub"""

    COMMAND = "`github`"
    ALIASES = "`gh`"
    FORGE = "GitHub"
    URL_FORMAT = (
        "https://raw.githubusercontent.com/{owner}/{repository}/{branch}/{file}"
    )


##############################################################################
class OpenFromGitLab(OpenFromForgeCommand):
    """Open a file from GitLab"""

    COMMAND = "`gitlab`"
    ALIASES = "`gl`"
    FORGE = "GitLab"
    URL_FORMAT = "https://gitlab.com/{owner}/{repository}/-/raw/{branch}/{file}"


### open_from_forge.py ends here



"""Test the history class."""

##############################################################################
# Python imports.
from typing import Sequence

##############################################################################
# Pytest imports.
from pytest import mark

##############################################################################
# Local imports.
from hike.support import History


##############################################################################
def test_empty_history_has_no_location() -> None:
    """An empty history object should have no location."""
    assert History().current_location is None


##############################################################################
def test_empty_history_has_no_item() -> None:
    """An empty history object should have no item."""
    assert History().current_item is None


##############################################################################
def test_empty_history_has_no_length() -> None:
    """An empty history object should have zero length."""
    assert len(History()) == 0


##############################################################################
def test_empty_history_has_nowhere_to_go() -> None:
    """An empty history should not be able to move."""
    assert History().can_go_backward is False
    assert History().can_go_forward is False


##############################################################################
@mark.parametrize("initial", ([1], [1, 2], range(100)))
def test_pre_populate_history_has_location(initial: Sequence[int]) -> None:
    """A pre-populated history should be at the last location."""
    assert History[int](initial).current_location == len(initial) - 1


##############################################################################
@mark.parametrize("initial", ([1], [1, 2], range(100)))
def test_pre_populate_history_has_an_item(initial: Sequence[int]) -> None:
    """A pre-populated history's item should be the last item."""
    assert History[int](initial).current_item == initial[-1]


##############################################################################
@mark.parametrize(
    "initial, expected",
    (
        ([1], 1),
        ([1, 2], 2),
        (range(100), 100),
        (range(200), 100),
    ),
)
def test_pre_populate_history_has_length(initial: Sequence[int], expected: int) -> None:
    """A pre-populated history should be the correct length."""
    assert len(History[int](initial, max_length=100)) == expected


##############################################################################
@mark.parametrize("values", ([1], [1, 2], range(100)))
def test_hand_populated_history_has_location(values: Sequence[int]) -> None:
    """A hand-populated history should end up at the last location."""
    history = History[int]()
    for value in values:
        history.add(value)
    assert history.current_location == len(values) - 1


##############################################################################
@mark.parametrize("values", ([1], [1, 2], range(100)))
def test_hand_populated_history_has_an_item(values: Sequence[int]) -> None:
    """A hand-populated history's item should be the last item."""
    history = History[int]()
    for value in values:
        history.add(value)
    assert history.current_item == values[-1]


##############################################################################
@mark.parametrize(
    "values, expected",
    (
        ([1], 1),
        ([1, 2], 2),
        (range(100), 100),
        (range(200), 100),
    ),
)
def test_hand_populated_history_has_length(
    values: Sequence[int], expected: int
) -> None:
    """A pre-populated history should be the correct length."""
    history = History[int](max_length=100)
    for value in values:
        history.add(value)
    assert len(history) == expected


##############################################################################
def test_backward() -> None:
    """We should be able to move backward through history until we can't."""
    history = History[int]([1, 2, 3])
    assert history.current_location == 2
    assert history.can_go_backward is True
    assert history.backward() is True
    assert history.current_location == 1
    assert history.can_go_backward is True
    assert history.backward() is True
    assert history.current_location == 0
    assert history.can_go_backward is False
    assert history.backward() is False
    assert history.current_location == 0


##############################################################################
def test_backward_empty() -> None:
    """We should not be able to go backward in an empty history."""
    history = History[None]([])
    assert history.current_location is None
    assert history.backward() is False
    assert history.current_location is None


##############################################################################
def test_forward() -> None:
    """We should be able to move forward through history until we can't."""
    history = History[int]([1, 2, 3])
    assert history.backward() is True
    assert history.backward() is True
    assert history.backward() is False
    assert history.can_go_backward is False
    assert history.current_location == 0
    assert history.can_go_forward is True
    assert history.forward() is True
    assert history.current_location == 1
    assert history.can_go_forward is True
    assert history.forward() is True
    assert history.current_location == 2
    assert history.can_go_forward is False
    assert history.forward() is False
    assert history.current_location == 2


##############################################################################
def test_forward_empty() -> None:
    """We should not be able to go forward in an empty history."""
    history = History[None]()
    assert history.current_location is None
    assert history.forward() is False
    assert history.current_location is None


##############################################################################
def test_iterator() -> None:
    """We should be able to use history as an iterator."""
    assert list(History[int]([1, 2, 3])) == [1, 2, 3]


##############################################################################
@mark.parametrize("desired, achieved", ((0, 0), (1, 1), (2, 2), (3, 2), (-1, 0)))
def test_goto(desired: int, achieved: int) -> None:
    """We should be able to go to specific locations in history."""
    assert History[None]((None, None, None)).goto(desired).current_location == achieved


##############################################################################
def test_goto_end() -> None:
    """We should be able to go to the end of the history."""
    history = History[None]((None,) * 100)
    assert history.goto(0).current_location == 0
    assert history.goto_end().current_location == 99


##############################################################################
def test_goto_end_of_nothing() -> None:
    """Going to the end when there is no history should cause no problem."""
    assert History[None]().goto_end().current_location is None


##############################################################################
def test_goto_empty() -> None:
    """Going to a location in an empty list should keep the location as `None`."""
    assert History[None]().goto(42).current_location is None


##############################################################################
def test_goto_and_movement_in_empty_history() -> None:
    """A combination of goto and movement in a empty history should be fine."""
    history = History[None]()
    assert history.current_location is None
    assert history.goto(42).current_location is None
    assert history.forward() is False
    assert history.current_location is None
    assert history.backward() is False
    assert history.current_location is None


##############################################################################
def test_delete_history() -> None:
    """We should be able to delete an item in history."""
    history = History[None]((None,))
    assert len(history) == 1
    del history[0]
    assert len(history) == 0


### test_history.py ends here



"""Messages to do with opening things."""

##############################################################################
# Python imports.
from dataclasses import dataclass
from pathlib import Path

##############################################################################
# httpx imports.
from httpx import URL, AsyncClient, HTTPStatusError, RequestError

##############################################################################
# Textual imports.
from textual.message import Message

##############################################################################
# Local imports.
from .. import USER_AGENT
from ..data import load_configuration
from ..types import HikeLocation


##############################################################################
@dataclass
class OpenFrom(Message):
    """Open a file starting at a particular location."""

    location: Path = Path(".")
    """The location to start browsing for a file."""


##############################################################################
@dataclass
class OpenLocation(Message):
    """Open a given location for viewing."""

    to_open: HikeLocation
    """The location to open."""


##############################################################################
@dataclass
class OpenFromHistory(Message):
    """Open a location found in history."""

    location: int
    """The location in the history to open."""


##############################################################################
@dataclass
class OpenFromForge(Message):
    """Open a file from a forge."""

    forge: str
    """The name of the forge."""

    raw_url_format: str
    """The format of the raw URL for the forge."""

    owner: str
    """The owner of the repository to open from."""

    repository: str
    """The repository to open the file from."""

    branch: str | None = None
    """The optional branch to open from."""

    filename: str | None = None
    """The optional name of the file to open."""

    async def url(self) -> URL | None:
        """The URL for the file on the forge.

        Returns:
            The URL if one could be worked out, or `None` if not.
        """
        filename = self.filename or "README.md"
        async with AsyncClient() as client:
            for candidate_branch in (
                [self.branch] if self.branch else load_configuration().main_branches
            ):
                url = self.raw_url_format.format(
                    owner=self.owner,
                    repository=self.repository,
                    branch=candidate_branch,
                    file=filename,
                )
                try:
                    response = await client.head(
                        url, follow_redirects=True, headers={"user-agent": USER_AGENT}
                    )
                except RequestError:
                    # A failed request would suggest further attempts aren't
                    # going to work so let's GTFO now.
                    return None
                try:
                    response.raise_for_status()
                except HTTPStatusError:
                    # Some sort of status error means we at least managed to
                    # contact the server, but didn't quite hit it with the
                    # URL we had; so let's go around again.
                    continue
                # If we're here we got a working URL, so let's go with that.
                return URL(url)
        return None


### opening.py ends here



"""Provides code for saving and loading the command history."""

##############################################################################
# Python imports.
from json import dumps, loads
from pathlib import Path

##############################################################################
# Local imports.
from ..types import CommandHistory
from .locations import data_dir


##############################################################################
def command_history_file() -> Path:
    """Get the path for the command history file.

    Returns:
        The path for the command history file.
    """
    return data_dir() / "command-history.json"


##############################################################################
def save_command_history(history: CommandHistory) -> None:
    """Save the command history to storage.

    Args:
        history: The command history to save.
    """
    command_history_file().write_text(
        dumps(list(history), indent=4),
        encoding="utf-8",
    )


##############################################################################
def load_command_history() -> CommandHistory:
    """Load the command history from storage.

    Returns:
        The loaded command history.
    """
    return CommandHistory(
        loads(history.read_text(encoding="utf-8"))
        if (history := command_history_file()).exists()
        else []
    )


### command_history.py ends here



"""Mouse support functions."""

##############################################################################
# Textual imports.
from textual.events import Click


##############################################################################
def is_copy_request_click(event: Click) -> bool:
    """Does the mouse click look like it's a copy request?

    Args:
        event: The mouse click event.

    Returns:
        `True` if the click was a copy request, `False` if not.
    """
    return (event.chain == 1 and event.ctrl) or (event.chain == 3 and not event.ctrl)


### mouse.py ends here



"""Provides a function for viewing things in the browser."""

##############################################################################
# Python imports.
from functools import singledispatch
from pathlib import Path
from webbrowser import open as open_url

##############################################################################
# httpx imports.
from httpx import URL


##############################################################################
@singledispatch
def view_in_browser(location: object) -> None:
    """View a location in the OS's web browser.

    Args:
        location: The location to view.
    """


@view_in_browser.register
def _(location: Path) -> None:
    open_url(f"file://{location}", new=2, autoraise=True)


@view_in_browser.register
def _(location: URL) -> None:
    open_url(str(location), new=2, autoraise=True)


### view_in_browser.py ends here



"""Provides application-wide command-oriented messages."""

##############################################################################
# Local imports.
from .main import (
    BookmarkLocation,
    ChangeCommandLineLocation,
    ChangeNavigationSide,
    CopyLocationToClipboard,
    CopyMarkdownToClipboard,
    Edit,
    JumpToCommandLine,
    JumpToDocument,
    Reload,
    SaveCopy,
    SearchBookmarks,
    ToggleNavigation,
)
from .navigation import (
    Backward,
    Forward,
    JumpToBookmarks,
    JumpToHistory,
    JumpToLocalBrowser,
    JumpToTableOfContents,
)

##############################################################################
# Exports.
__all__ = [
    "Backward",
    "BookmarkLocation",
    "ChangeCommandLineLocation",
    "ChangeNavigationSide",
    "CopyLocationToClipboard",
    "CopyMarkdownToClipboard",
    "Edit",
    "Forward",
    "JumpToBookmarks",
    "JumpToCommandLine",
    "JumpToDocument",
    "JumpToHistory",
    "JumpToLocalBrowser",
    "JumpToTableOfContents",
    "Reload",
    "SaveCopy",
    "SearchBookmarks",
    "ToggleNavigation",
]

### __init__.py ends here



"""Provides the bookmarks for the navigation panel."""

##############################################################################
# Python imports.
from dataclasses import dataclass
from typing import cast

##############################################################################
# httpx imports.
from httpx import URL

##############################################################################
# Rich imports.
from rich.text import Text

##############################################################################
# Textual imports.
from textual import on, work
from textual.message import Message
from textual.widgets.option_list import Option

##############################################################################
# Textual enhanced imports.
from textual_enhanced.binding import HelpfulBinding
from textual_enhanced.dialogs import Confirm, ModalInput
from textual_enhanced.widgets import EnhancedOptionList

##############################################################################
# Local imports.
from ...data import Bookmark, Bookmarks
from ...messages import OpenLocation


##############################################################################
class BookmarkView(Option):
    """The view of a bookmark."""

    def __init__(self, bookmark: Bookmark) -> None:
        """Initialise the object.

        Args:
            bookmark: The bookmark.
        """
        self.bookmark = bookmark
        """The bookmark."""
        super().__init__(
            Text.from_markup(
                f":{'globe_with_meridians' if isinstance(bookmark.location, URL) else 'page_facing_up'}: "
                f"[bold]{bookmark.title}[/]\n[dim]{bookmark.location}[/]",
                overflow="ellipsis",
            ),
            id=str(bookmark.location),
        )


##############################################################################
class BookmarksView(EnhancedOptionList):
    """The display of bookmarks."""

    DEFAULT_CSS = """
    BookmarksView {
        height: 1fr;
        border: none;
        &:focus {
            border: none;
        }
    }
    """

    HELP = """
    ## Bookmarks

    This panel contains your bookmarks. Here you can visit them and manage them.
    """

    BINDINGS = [
        HelpfulBinding(
            "r",
            "rename",
            "Rename",
            show=False,
            tooltip="Rename the highlighted bookmark",
        ),
        HelpfulBinding(
            "delete",
            "delete",
            "Delete",
            show=False,
            tooltip="Delete the highlighted bookmark",
        ),
    ]

    def update(self, bookmarks: Bookmarks) -> None:
        """Up[date the content of the bookmarks view.

        Args:
            bookmarks: The bookmarks to show.
        """
        with self.preserved_highlight:
            self.clear_options().add_options(
                BookmarkView(bookmark) for bookmark in sorted(bookmarks)
            )

    @on(EnhancedOptionList.OptionSelected)
    def _visit_bookmark(self, message: EnhancedOptionList.OptionSelected) -> None:
        """Visit the current bookmark.

        Args:
            message: The message requesting the visit.
        """
        message.stop()
        assert isinstance(message.option, BookmarkView)
        self.post_message(OpenLocation(message.option.bookmark.location))

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        """Check if an action is possible to perform right now.

        Args:
            action: The action to perform.
            parameters: The parameters of the action.

        Returns:
            `True` if it can perform, `False` or `None` if not.
        """
        if action in ("edit", "delete"):
            return bool(self.option_count) and self.highlighted is not None
        return True

    @property
    def current_bookmark(self) -> Bookmark | None:
        """The currently-highlighted bookmark."""
        return (
            None
            if self.highlighted is None
            else cast(BookmarkView, self.get_option_at_index(self.highlighted)).bookmark
        )

    @dataclass
    class Renamed(Message):
        """Message sent when a bookmark is renamed."""

        renamed_from: Bookmark
        """The bookmark that was being renamed."""
        renamed_to: Bookmark
        """The renamed bookmark."""

    @work
    async def action_rename(self) -> None:
        """Rename the currently-highlighted bookmark."""
        if self.current_bookmark is None:
            return
        if title := await self.app.push_screen_wait(
            ModalInput(
                "The title for the bookmark", initial=self.current_bookmark.title
            )
        ):
            self.post_message(
                self.Renamed(
                    self.current_bookmark,
                    Bookmark(title, self.current_bookmark.location),
                )
            )

    @dataclass
    class Deleted(Message):
        """Message sent when a bookmark is deleted."""

        deleted: Bookmark
        """The bookmark that was deleted."""

    @work
    async def action_delete(self) -> None:
        """Delete the currently-highlighted bookmark."""
        if self.current_bookmark is None:
            return
        if await self.app.push_screen_wait(
            Confirm("Delete bookmark?", "Are you sure you want to delete the bookmark?")
        ):
            self.post_message(self.Deleted(self.current_bookmark))


### bookmarks_view.py ends here



"""Provides a widget for browsing the local filesystem."""

##############################################################################
# Python imports.
from pathlib import Path
from typing import Iterable

##############################################################################
# Textual imports.
from textual import on
from textual.widgets import DirectoryTree

##############################################################################
# Local imports.
from ...data import maybe_markdown
from ...messages import OpenLocation


##############################################################################
class LocalView(DirectoryTree):
    """A browser for the local filesystem."""

    def filter_paths(self, paths: Iterable[Path]) -> Iterable[Path]:
        """Filter a directory for things that look like Markdown.

        Args:
            paths: The paths to filter down.

        Returns:
            The paths filtered as useful for Markdown viewing.
        """
        return (
            path
            for path in paths
            if (path.is_dir() and not path.name.startswith("."))
            or (path.is_file() and maybe_markdown(path))
        )

    @on(DirectoryTree.FileSelected)
    def view_file(self, message: DirectoryTree.FileSelected) -> None:
        """View the selected file.

        Args:
            message: The message requesting that the file be viewed.
        """
        message.stop()
        self.post_message(OpenLocation(message.path))


### local_view.py ends here



"""Tests for the location type testing code."""

##############################################################################
# Python imports.
from pathlib import Path

##############################################################################
# httpx imports.
from httpx import URL

##############################################################################
# Pytest imports.
from pytest import mark

##############################################################################
# Local imports.
from hike.data import looks_urllike, maybe_markdown
from hike.types import HikeLocation


##############################################################################
@mark.parametrize(
    "to_test, expected",
    (
        (Path("test.md"), True),
        (Path("test.markdown"), True),
        (Path("test.txt"), False),
        (URL("http://example.com/test.md"), True),
        (URL("http://example.com/test.markdown"), True),
        (URL("http://example.com/test.txt"), False),
        (URL("https://example.com/test.md"), True),
        (URL("https://example.com/test.markdown"), True),
        (URL("https://example.com/test.txt"), False),
    ),
)
def test_maybe_markdown(to_test: HikeLocation, expected: bool) -> None:
    """We should be able to make a good guess at what's a Markdown location."""
    assert maybe_markdown(to_test) is expected


##############################################################################
@mark.parametrize(
    "to_test, expected",
    (
        ("http://example.com/", True),
        ("https://example.com/", True),
        ("http://example", True),
        ("https://example", True),
        ("example.com/", False),
        ("example.com/", False),
        ("", False),
        ("http", False),
        ("https", False),
        ("http://", False),
        ("https://", False),
    ),
)
def test_looks_urllike(to_test: str, expected: bool) -> None:
    """We should be able to make a good guess at what's a URL."""
    assert looks_urllike(to_test) is expected


### test_location_types.py ends here



"""Functions for testing location types."""

##############################################################################
# Python imports.
from functools import singledispatch
from pathlib import Path

##############################################################################
# httpx imports.
from httpx import URL

##############################################################################
from typing_extensions import TypeIs

##############################################################################
# Local imports.
from ..types import HikeLocation
from .config import load_configuration


##############################################################################
@singledispatch
def maybe_markdown(location: object) -> bool:
    """Does the given location look like it might be a Markdown file?

    Args:
        location: The location to test.

    Returns:
        `True` if the location looks like it's Markdown, `False` if not.
    """
    return False


##############################################################################
@maybe_markdown.register
def _(location: Path) -> bool:
    return location.suffix.lower() in load_configuration().markdown_extensions


##############################################################################
@maybe_markdown.register
def _(location: str) -> bool:
    return maybe_markdown(Path(location))


##############################################################################
@maybe_markdown.register
def _(location: URL) -> bool:
    return maybe_markdown(location.path)


##############################################################################
def looks_urllike(candidate: str) -> bool:
    """Does the given value look like a URL?

    Args:
        candiate: The candidate to test.

    Returns:
        `True` if the string looks like a URL, `False` if not.
    """
    return (url := URL(candidate)).is_absolute_url and url.scheme in ("http", "https")


##############################################################################
def is_editable(location: HikeLocation) -> TypeIs[Path]:
    """Is the given location one that can be edited?

    Args:
        location: The location to test.

    Returns:
        `True` if the location can be edited, `False` if not.
    """
    return isinstance(location, Path)


### location_types.py ends here
